package config

import (
	"errors"
	"fmt"
	"io/ioutil"
	"os"
	"path"
	"strings"

	"coopcloud.tech/abra/client"
)

// Type aliases to make code hints easier to understand

// AppEnv is a map of the values in an apps env config
type AppEnv = map[string]string

// AppName is AppName
type AppName = string

// AppFile represents app env files on disk without reading the contents
type AppFile struct {
	Path   string
	Server string
}

// AppFiles is a slice of appfiles
type AppFiles map[AppName]AppFile

// App reprents an app with its env file read into memory
type App struct {
	Name   AppName
	Type   string
	Domain string
	Env    AppEnv
	File   AppFile
}

// StackName gets what the docker safe stack name is for the app
func (a App) StackName() string {
	return SanitiseAppName(a.Name)
}

// SORTING TYPES

// ByServer sort a slice of Apps
type ByServer []App

func (a ByServer) Len() int      { return len(a) }
func (a ByServer) Swap(i, j int) { a[i], a[j] = a[j], a[i] }
func (a ByServer) Less(i, j int) bool {
	return strings.ToLower(a[i].File.Server) < strings.ToLower(a[j].File.Server)
}

// ByServerAndType sort a slice of Apps
type ByServerAndType []App

func (a ByServerAndType) Len() int      { return len(a) }
func (a ByServerAndType) Swap(i, j int) { a[i], a[j] = a[j], a[i] }
func (a ByServerAndType) Less(i, j int) bool {
	if a[i].File.Server == a[j].File.Server {
		return strings.ToLower(a[i].Type) < strings.ToLower(a[j].Type)
	}
	return strings.ToLower(a[i].File.Server) < strings.ToLower(a[j].File.Server)
}

// ByType sort a slice of Apps
type ByType []App

func (a ByType) Len() int      { return len(a) }
func (a ByType) Swap(i, j int) { a[i], a[j] = a[j], a[i] }
func (a ByType) Less(i, j int) bool {
	return strings.ToLower(a[i].Type) < strings.ToLower(a[j].Type)
}

// ByName sort a slice of Apps
type ByName []App

func (a ByName) Len() int      { return len(a) }
func (a ByName) Swap(i, j int) { a[i], a[j] = a[j], a[i] }
func (a ByName) Less(i, j int) bool {
	return strings.ToLower(a[i].Name) < strings.ToLower(a[j].Name)
}

func readAppEnvFile(appFile AppFile, name AppName) (App, error) {
	env, err := ReadEnv(appFile.Path)
	if err != nil {
		return App{}, fmt.Errorf("env file for '%s' couldn't be read: %s", name, err.Error())
	}
	app, err := newApp(env, name, appFile)
	if err != nil {
		return App{}, fmt.Errorf("env file for '%s' has issues: %s", name, err.Error())
	}
	return app, nil
}

// newApp creates new App object
func newApp(env AppEnv, name string, appFile AppFile) (App, error) {
	// Checking for type as it is required - apps wont work without it
	domain := env["DOMAIN"]
	apptype, ok := env["TYPE"]
	if !ok {
		return App{}, errors.New("missing TYPE variable")
	}
	return App{
		Name:   name,
		Domain: domain,
		Type:   apptype,
		Env:    env,
		File:   appFile,
	}, nil
}

// LoadAppFiles gets all app files for a given set of servers or all servers
func LoadAppFiles(servers ...string) (AppFiles, error) {
	appFiles := make(AppFiles)
	if len(servers) == 1 {
		if servers[0] == "" {
			// Empty servers flag, one string will always be passed
			var err error
			servers, err = getAllFoldersInDirectory(ABRA_SERVER_FOLDER)
			if err != nil {
				return nil, err
			}
		}
	}
	for _, server := range servers {
		serverDir := path.Join(ABRA_SERVER_FOLDER, server)
		files, err := getAllFilesInDirectory(serverDir)
		if err != nil {
			return nil, err
		}
		for _, file := range files {
			appName := strings.TrimSuffix(file.Name(), ".env")
			appFilePath := path.Join(ABRA_SERVER_FOLDER, server, file.Name())
			appFiles[appName] = AppFile{
				Path:   appFilePath,
				Server: server,
			}
		}
	}
	return appFiles, nil
}

// GetApp loads an apps settings, reading it from file, in preparation to use it
//
// ONLY use when ready to use the env file to keep IO down
func GetApp(apps AppFiles, name AppName) (App, error) {
	appFile, exists := apps[name]
	if !exists {
		return App{}, fmt.Errorf("cannot find app file with name '%s'", name)
	}
	app, err := readAppEnvFile(appFile, name)
	if err != nil {
		return App{}, err
	}
	return app, nil
}

// GetApps returns a slice of Apps with their env files read from a given slice of AppFiles
func GetApps(appFiles AppFiles) ([]App, error) {
	var apps []App
	for name := range appFiles {
		app, err := GetApp(appFiles, name)
		if err != nil {
			return nil, err
		}
		apps = append(apps, app)
	}
	return apps, nil
}

// CopyAppEnvSample copies the example env file for the app into the users env files
func CopyAppEnvSample(appType, appName, server string) error {
	envSamplePath := path.Join(ABRA_DIR, "apps", appType, ".env.sample")
	envSample, err := ioutil.ReadFile(envSamplePath)
	if err != nil {
		return err
	}

	appEnvPath := path.Join(ABRA_DIR, "servers", server, fmt.Sprintf("%s.env", appName))
	if _, err := os.Stat(appEnvPath); err == nil {
		return fmt.Errorf("%s already exists?", appEnvPath)
	}

	err = ioutil.WriteFile(appEnvPath, envSample, 0755)
	if err != nil {
		return err
	}

	return nil
}

// SanitiseAppName makes a app name usable with Docker by replacing illegal characters
func SanitiseAppName(name string) string {
	return strings.ReplaceAll(name, ".", "_")
}

// GetAppStatuses queries servers to check the deployment status of given apps
func GetAppStatuses(appFiles AppFiles) (map[string]string, error) {
	servers := appFiles.GetServers()
	ch := make(chan client.StackStatus, len(servers))
	for _, server := range servers {
		go func(s string) {
			ch <- client.QueryStackStatus(s)
		}(server)
	}

	statuses := map[string]string{}
	for range servers {
		status := <-ch
		for _, service := range status.Services {
			name := service.Spec.Labels[client.StackNamespace]
			if _, ok := statuses[name]; !ok {
				statuses[name] = "deployed"
			}
		}
	}

	return statuses, nil
}
